clear all;
close all;

rng(1)
% 2-dim toy data generation
C=5;
samples_per_class=100;
sigma=1.0;
num_samples = C*samples_per_class;

% classes A-E with negative labels
noise_dim = 100;
feature_dim = 2;
dim = noise_dim+feature_dim;

cluster_centers = [2  -4; -3  3; 3   1; -5 -2; -4 -5];
colors_samples = {'og','sr','xb','+m','dy'};
colors_classifiers = {'g','r','b','m','y'};

data = zeros(dim+1, num_samples);
for j=1:C
    data(:, (j-1)*samples_per_class+1:j*samples_per_class) = ...
        [repmat(cluster_centers(j, :),samples_per_class,1) + randn(samples_per_class,2)*sigma ...
        randn(samples_per_class, noise_dim)*3 ...
        j*ones(samples_per_class, 1)]';
end

% Suffle data
shuffle = randperm(size(data, 2));
data = data(:, shuffle);

% Split data into training and test set
N = num_samples-0.1*num_samples;
train_data = data(:, 1:N);
test_data = data(:, N+1:end);

figure;
subplot(1,2,1);
hold on
for j=1:C
   scatter(train_data(1, train_data(end, :) == j),...
   train_data(2, train_data(end, :) == j),colors_samples{j});
end
axis equal;
title('Training Data');

subplot(1,2,2);
hold on
for j=1:C
   scatter(test_data(1, test_data(end, :) == j),...
   test_data(2, test_data(end, :) == j),colors_samples{j});
end
axis equal;
title('Test Data');


X = train_data(1:dim, :)';
%Normalize data
X = X/max(X(:));
y = train_data(end, :)';

%%
lambda1 = 0.01;
lambda2 = 0.02;

L = norm(X,'fro')^2/N + lambda1;
tau = 1/L;


max_iter=2000;



% compute classifier scores
classifier = @(W, b, X) X*W + repmat(b, size(X, 1), 1);

% evaluate loss given classifier scores
scores = @(classifier) exp(classifier) ./ repmat(sum(exp(classifier), 2), 1, C);
loss = @(scores) sum(-log(scores(sub2ind(size(scores), (1:N)', y))), 1);

objective = @(W, b) loss(scores(classifier(W, b, X)))/N ...
    + 0.5 * lambda1 * sum(W(:).^2)...
    + 0.5 * lambda1 * sum(b(:).^2);...
%    + lambda2 * sum(W(:));

grad_W = @(W, scores, j) ...
    sum(X.*repmat(scores(:, j) - double(y == j), 1, dim), 1)'/N  ...
    + lambda1 * W(:, j);

grad_b = @(b, scores, j) ...
    sum(scores(:, j) - double(y == j), 1)'/N  ...
    + lambda1 * b(j);


W = rand(dim, C);
b = zeros(1, C);


obj_val = zeros(1, max_iter);
test_error = zeros(1, max_iter);

% func = @(x) fun(x, dim, C, X, scores, classifier, objective, grad_W, grad_b);
% 
% options = optimoptions('fminunc','GradObj','on','DerivativeCheck','on');
% res = fminunc(func, [W(:);b(:)], options);
% W = reshape(res(1:dim*C), dim, C);
% b = reshape(res(dim*C+1:end), 1, C);


for iter=1:max_iter
    scores_train = scores(classifier(W, b, X));
    scores_test = scores(classifier(W, b, test_data(1:dim,:)'));
    
    [~, argmin] = max(scores_test');
    
    obj_val(iter) = objective(W, b);
    test_error(iter) = nnz(argmin'~=test_data(end,:)')/size(test_data,2)*100;
    fprintf('It=%d   obj=%f   test_error=%2.2f%%\n', ...
        iter, obj_val(iter), test_error(iter));
    
    for j=1:C   
        W(:, j) = W(:, j) - tau*grad_W(W, scores_train, j);
        b(j) = b(j) - tau*grad_b(b, scores_train, j);
    end
    
    % prox step 
    W = max(abs(W)-tau*lambda2, 0).*sign(W);

    
    % stopping 
    if(iter > 1 && obj_val(iter-1) - obj_val(iter) < 1e-12)
        break;
    end
end

%% Plot objective value and test error
figure;
subplot(1,2,1);
plot(obj_val(1:iter));
title('Objective value');
subplot(1,2,2);
plot(test_error(1:iter));
title('Test error');


%% Plot classifiers on test data
s = 0.1;
[x1Grid, x2Grid] = meshgrid(-4:s:4,...
    -4:s:4);
xGrid = [x1Grid(:),x2Grid(:)];
scores_train = classifier(W(1:2,:), b, xGrid);
figure;
hold on
for j=1:C
    scatter(test_data(1, test_data(end, :)==j),...
        test_data(2, test_data(end, :)==j),colors_samples{j});
    
    [~,h] = contour(x1Grid,x2Grid,reshape(scores_train(:, j),size(x1Grid)),[0 0],'k');
    h.LineColor = colors_classifiers{j};   
end
drawnow
title('Classifiers on Test data')