clear all;
close all;

rng(1)

images = loadMNISTImages('t10k-images-idx3-ubyte');
labels = loadMNISTLabels('t10k-labels-idx1-ubyte')+1;
test_data = [images; labels'];

images = loadMNISTImages('train-images-idx3-ubyte');
[dim, N] = size(images);
labels = loadMNISTLabels('train-labels-idx1-ubyte')+1;
C=max(labels);
train_data = [images; labels'];
X = train_data(1:dim, :)';
%Normalize data
%X = X/max(X(:));

y = train_data(end, :)';

%%
lambda1 = 0.001;
lambda2 = 0.002;

L = norm(X,'fro')^2/N + lambda1;
%tau = 1/L;
tau = 0.7;


max_iter=2000;



% compute classifier scores
classifier = @(W, b, X) X*W + repmat(b, size(X, 1), 1);

% evaluate loss given classifier scores
scores = @(classifier) exp(classifier) ./ repmat(sum(exp(classifier), 2), 1, C);
loss = @(scores) sum(-log(scores(sub2ind(size(scores), (1:N)', y))), 1);

objective = @(W, b) loss(scores(classifier(W, b, X)))/N ...
    + 0.5 * lambda1 * sum(W(:).^2)...
    + 0.5 * lambda1 * sum(b(:).^2)...
    + lambda2 * sum(W(:));

grad_W = @(W, scores, j) ...
    sum(X.*repmat(scores(:, j) - double(y == j), 1, dim), 1)'/N  ...
    + lambda1 * W(:, j);

grad_b = @(b, scores, j) ...
    sum(scores(:, j) - double(y == j), 1)'/N  ...
    + lambda1 * b(j);


W = ones(dim, C);
b = zeros(1, C);


obj_val = zeros(1, max_iter);
test_error = zeros(1, max_iter);

% func = @(x) fun(x, dim, C, X, scores, classifier, objective, grad_W, grad_b);
% 
% options = optimoptions('fminunc','GradObj','on','DerivativeCheck','on');
% res = fminunc(func, [W(:);b(:)], options);
% W = reshape(res(1:dim*C), dim, C);
% b = reshape(res(dim*C+1:end), 1, C);


for iter=1:max_iter
    scores_train = scores(classifier(W, b, X));
    scores_test = scores(classifier(W, b, test_data(1:dim,:)'));
    
    [~, argmin] = max(scores_test');
    
    obj_val(iter) = objective(W, b);
    test_error(iter) = nnz(argmin'~=test_data(end,:)')/size(test_data,2)*100;
    fprintf('It=%d   obj=%f   test_error=%2.2f%%\n', ...
        iter, obj_val(iter), test_error(iter));
    
    for j=1:C   
        W(:, j) = W(:, j) - tau*grad_W(W, scores_train, j);
        b(j) = b(j) - tau*grad_b(b, scores_train, j);
    end
    
    W = max(abs(W)-tau*lambda2, 0).*sign(W);

    %if(iter > 1 && obj_val(iter-1) - obj_val(iter) < 1e-12)
    %    break;
    %end
end

%% Plot objective value and test error
figure;
subplot(1,2,1);
plot(obj_val(1:iter));
title('Objective value');
subplot(1,2,2);
plot(test_error(1:iter));
title('Test error');

